/********************************************************************
*
* SUBDUE
*
* FILE NAME: cluster.c
* VERSION:   4.2.d.2
*
********************************************************************/


#include "subdue.h"
#include "cluster.h"



/*******************************************************************************
FUNCTION NAME: CL_CreateLattice
INPUTS:			none
RETURNS:		PCLASSIFICATION_LATTICE
PURPOSE:		Create and return the classification lattice root node.
CALLED BY:		main.c: main()
*******************************************************************************/
PCLASSIFICATION_LATTICE CL_CreateLattice()
{
	PCLASSIFICATION_LATTICE newCLRoot;
	char labelStr[MAX_TOKEN_LENGTH];
	
#ifdef DEBUG_TRACE
	printf( "%s: CL_CreateLattice()\n", __FILE__ );
#endif
	
	newCLRoot = (PCLASSIFICATION_LATTICE) Malloc( sizeof(CLASSIFICATION_LATTICE) );
	
	sprintf(labelStr, "ROOT" );									// generating sub label
	newCLRoot->subLabel = (char *)Malloc( strlen(labelStr) + 1);
	strcpy(newCLRoot->subLabel, labelStr);
	newCLRoot->trueLabel = (char *)Malloc( strlen(labelStr) + 1 );
	strcpy(newCLRoot->trueLabel, labelStr);


//	newCLRoot->definition = NULL;								// initializing members
//	newCLRoot->instances = NULL;
	newCLRoot->children = NULL;
	newCLRoot->numberOfChildren = 0;
	newCLRoot->numberOfParents = 1;								// root has dummy parent
																// this is used during destroy
	newCLRoot->referenceCount = 0;

	return newCLRoot;
}

/*******************************************************************************
FUNCTION NAME: CL_CreateNode
INPUTS:			PGRAPH_VARIABLES GV
				PSUB sub
RETURNS:		PCLASSIFICATION_LATTICE
PURPOSE:		Create and return a classification lattice (CL) node.
CALLED BY:		main.c: main()
*******************************************************************************/
PCLASSIFICATION_LATTICE CL_CreateNode(PGRAPH_VARIABLES GV, PSUB sub)
{
	PCLASSIFICATION_LATTICE newCLNode;
	char labelStr[12000];			// this sould be really big, since it has to hold
									// potentially many node and edge labels
//	PSUB_GRAPH definition;
//	ULONG i;
	
#ifdef DEBUG_TRACE
	printf( "%s: CL_CreateNode()\n", __FILE__ );
#endif
	
	newCLNode = (PCLASSIFICATION_LATTICE) Malloc( sizeof(CLASSIFICATION_LATTICE) );

	// create label of form "Sub_n"
	sprintf(labelStr, "%s%u\0", SUB_LABEL_PREFIX, GV->subNumber );	// generating sub label
	newCLNode->subLabel = (char *)Malloc( strlen(labelStr) + 1 );
	strcpy(newCLNode->subLabel, labelStr);

	// create label that describes the sub in words
	if (GV->trueLabel) {
		labelStr[0] = '\0';
		CL_MakeTrueLabel( GV, sub->definition, labelStr);			// generating sub label
		newCLNode->trueLabel = (char *)Malloc( strlen(labelStr) + 1 );
		strcpy(newCLNode->trueLabel, labelStr);
	}
	else 
		newCLNode->trueLabel = NULL;

//	newCLNode->instances = NULL;								// initializing members
	newCLNode->children = NULL;
	newCLNode->numberOfChildren = 0;
	newCLNode->numberOfParents = 0;
	newCLNode->referenceCount = 0;

/*	newCLNode->definition = (PSUB_GRAPH) Malloc(sizeof(SUB_GRAPH));// allocate space for definition

	definition = newCLNode->definition;
	definition->numberOfVertices = sub->definition->numberOfVertices;
	definition->numberOfEdges = sub->definition->numberOfEdges;
	definition->matchingOrder = NULL;
	definition->matchCost = 0;
	definition->refCount = 0;

	//** this may not work. depends on how the graph is constructed and retained after 
	//   a compression. -\/-
	// Aslo, it may not be needed.

	// copy the vertices
	definition->vertices = (PSUB_GRAPH_VERTEX) 
		Malloc(definition->numberOfVertices * sizeof(SUB_GRAPH_VERTEX));

	for (i=0; i < definition->numberOfVertices; i++) {
		definition->vertices[i] = sub->definition->vertices[i];
	}

	// copy the vertices
	definition->edges = (PSUB_GRAPH_EDGE) 
		Malloc(definition->numberOfEdges * sizeof(SUB_GRAPH_EDGE));

	for (i=0; i < definition->numberOfEdges; i++) {
		definition->edges[i] = sub->definition->edges[i];
	}
*/

	return newCLNode;
}
/*******************************************************************************
FUNCTION NAME: CL_Insert
INPUTS:			PGRAPH_VARIABLES GV
				PCLASSIFICATION_LATTICE clRoot
				PSUB sub
RETURNS:		BOOLEAN
PURPOSE:		Insert a sub into CL rooted at clRoot.
CALLED BY:		main.c: main()
*******************************************************************************/
BOOLEAN CL_Insert(PGRAPH_VARIABLES GV, PCLASSIFICATION_LATTICE clRoot, PSUB sub)
{
	PCLASSIFICATION_LATTICE clNode;
	PCLASSIFICATION_LATTICE clParent;
	PSUB_GRAPH_VERTEX vertices;
	PSUB_GRAPH_VERTEX currentVertex;
	PGRAPH_VERTEX graphVertices;
	PGRAPH_VERTEX currentGraphVertex;
	LABEL label;
	ULONG parentCount;
	ULONG i;
	
#ifdef DEBUG_TRACE
	printf( "%s: CL_Insert()\n", __FILE__ );
#endif

	clNode = CL_CreateNode(GV, sub);					// create new CL node based on sub
	vertices = sub->definition->vertices;				// get vertices
	parentCount = 0;

	for (i=0; i < sub->definition->numberOfVertices; i++) {	// for all the vertices of the new node
		graphVertices = GV->graph->vertices;
		currentVertex = &vertices[i];
		currentGraphVertex = &graphVertices[currentVertex->indexInGraph];
		label = GV->labelList[currentGraphVertex->labelIndex];
		
		if ( label.labelType == 's' ) {							 // for string labels only
			clParent = CL_Find(clRoot, label.content.stringValue );	// find CL node

			if (clParent) {											// if found
				CL_HookUp(clParent, clNode);							// hook up
				parentCount++;
			}
		}
	}

	if (parentCount == 0) {								// if node was not hooked up anywhere
		CL_HookUp(clRoot, clNode);							// hook it to the root
	}

	return TRUE;
}

/*******************************************************************************
FUNCTION NAME: CL_HookUp
INPUTS:			PCLASSIFICATION_LATTICE clParent
				PCLASSIFICATION_LATTICE clChild
RETURNS:		BOOLEAN
PURPOSE:		Hook up CL node as child of clParent.
CALLED BY:		cluster.c: CL_Insert()
*******************************************************************************/
BOOLEAN CL_HookUp(PCLASSIFICATION_LATTICE clParent, PCLASSIFICATION_LATTICE clChild)
{
	
#ifdef DEBUG_TRACE
	printf( "%s: CL_HookUp()\n", __FILE__ );
#endif
	
	clParent->numberOfChildren++;								// increase number of children
	clParent->children = (PCLASSIFICATION_LATTICE *) Realloc(clParent->children, 
		clParent->numberOfChildren * sizeof(PCLASSIFICATION_LATTICE));

	if (clParent->children == NULL) 
		return FALSE;											// hook up fails (no memory)

	clParent->children[clParent->numberOfChildren - 1] = clChild;
	clChild->numberOfParents++;									// increasing number of parents

	return TRUE;
}

/*******************************************************************************
FUNCTION NAME: CL_Find
INPUTS:			PCLASSIFICATION_LATTICE clRoot
				const char *label
RETURNS:		PCLASSIFICATION_LATTICE
PURPOSE:		Find a CL node in a CL pointed to by root.
CALLED BY:		cluster.c: CL_Insert()
*******************************************************************************/
PCLASSIFICATION_LATTICE CL_Find(PCLASSIFICATION_LATTICE clRoot, const char *label)
{
	PCLASSIFICATION_LATTICE CLNode;
	ULONG i;
	
#ifdef DEBUG_TRACE
	printf( "%s: CL_Find()\n", __FILE__ );
#endif

	if (strcmp(label, clRoot->subLabel) == 0) {					// if this is it
		return clRoot;
	}

	for (i=0; i < clRoot->numberOfChildren; i++) {				// for all the children
		CLNode = CL_Find(clRoot->children[i], label);			// find children recursively
		if (CLNode)												// if child retured not NULL
			return CLNode;										// we got it
	}

	return NULL;												// this branch does not have 'label'
}

/*******************************************************************************
FUNCTION NAME: CL_Print
INPUTS:			PGRAPH_VARIABLES GV
				PCLASSIFICATION_LATTICE clNode
				ULONG height
RETURNS:		none
PURPOSE:		Print a CL preorder style.
CALLED BY:		main.c: main()
*******************************************************************************/
void CL_Print(PGRAPH_VARIABLES GV, PCLASSIFICATION_LATTICE clNode, ULONG height)
{
	int doFile;
	ULONG i;
		
#ifdef DEBUG_TRACE
	printf( "%s: CL_Print()\n", __FILE__ );
#endif

	clNode->referenceCount++;
	doFile = (clNode->referenceCount == clNode->numberOfParents) ? 1 : 0;
	
	// Printing lattice on screen
	for (i=0; i < height; i++)						// create appopriate tabulation
		printf("   ");

	printf("%s\n", clNode->subLabel);				// print sub label on screen
	
	// Printing graph information into graph file
	if (GV->clusterFile && doFile) {							// if graph file is open
		fprintf(GV->clusterFile, "%s;\n ", clNode->subLabel);		// print node label
		fprintf(GV->clusterFile, "\t%s [color=%s",				// select color
			clNode->subLabel, ATT_COLOR[height % ATT_COLOR_COUNT]);
		if (GV->trueLabel)
			fprintf(GV->clusterFile, ", label=\"%s\"", clNode->trueLabel);
		fprintf(GV->clusterFile, "];\n");
	}

	for (i=0; i < clNode->numberOfChildren; i++) {		// for all the children
		if (GV->clusterFile && doFile) {						// if file is open
			fprintf(GV->clusterFile, "\t%s -> ", clNode->subLabel);
		}
		CL_Print(GV, clNode->children[i], height + 1);		// print children recursively
	}
}

/*******************************************************************************
FUNCTION NAME: CL_Destroy
INPUTS:			PCLASSIFICATION_LATTICE clNode
RETURNS:		none
PURPOSE:		Deallocate a CL (or a CL node and all its children).
CALLED BY:		main.c: main()
*******************************************************************************/
void CL_Destroy(PCLASSIFICATION_LATTICE clNode)
{
	ULONG i;
		
#ifdef DEBUG_TRACE
	printf( "%s: CL_Destroy()\n", __FILE__ );
#endif
	
	clNode->numberOfParents--;
	if ( clNode->numberOfParents == 0 ) {				// if no more parents

/*		if (clNode->definition)
			DestroySubGraph(clNode->definition);
*/
		for (i=0; i < clNode->numberOfChildren; i++) {		// for all the children
			CL_Destroy(clNode->children[i]);				// destroy children recursively
		}

		Free(clNode->children);
		Free(clNode->subLabel);
		Free(clNode->trueLabel);
		Free(clNode);										// deallocate this node
	}
}

/*******************************************************************************
FUNCTION NAME: CL_Finish
INPUTS:			PGRAPH_VARIABLES GV
				PCLASSIFICATION_LATTICE clNode
RETURNS:		none
PURPOSE:		Deallocate a CL (or a CL node and all its children).
CALLED BY:		main.c: main()
*******************************************************************************/
void CL_Finish(PGRAPH_VARIABLES GV, PCLASSIFICATION_LATTICE clNode)
{

#ifdef DEBUG_TRACE
	printf( "%s: CL_Finish()\n", __FILE__ );
#endif
	
	if (GV->cluster) {				// if clustering
		printf("Classification lattice:\n");
		CL_Print(GV, clNode, 0);		// printing lattice

		fprintf(GV->clusterFile, "}\n");		// close graph body
		fclose(GV->clusterFile);				// close cluster file

		CL_Destroy(clNode);				// destroy classification lattice
	}
}

/*******************************************************************************
FUNCTION NAME: CL_MakeTrueLabel
INPUTS:		PGRAPH_VARIABLES GV, 
			PSUB_GRAPH subGraph, 
			char *trueLabel
RETURNS:	void
PURPOSE:	Make a label that describes the node by descriptions of edges and
			vertices.
CALLED BY:	cluster.c: CL_CreateNode() 
*******************************************************************************/

void CL_MakeTrueLabel( PGRAPH_VARIABLES GV, PSUB_GRAPH subGraph, char *trueLabel )
{
	ULONG vertexIndex;
	ULONG edgeIndex;
	PSUB_GRAPH_VERTEX vertices;
	PSUB_GRAPH_VERTEX currentVertex;
	PSUB_GRAPH_EDGE currentSubEdges;
	PGRAPH_VERTEX graphVertices;
	PGRAPH_VERTEX currentGraphVertex;
	char labelBuffer[MAX_TOKEN_LENGTH];

	
#ifdef DEBUG_TRACE
	printf( "%s: CL_MakeTrueLabel()\n", __FILE__ );
#endif
	
	vertices = subGraph->vertices;
	graphVertices = GV->graph->vertices;

	for ( vertexIndex = 0; vertexIndex < subGraph->numberOfVertices; vertexIndex++ )
	{
		currentVertex = &vertices[vertexIndex];
		currentGraphVertex = &graphVertices[currentVertex->indexInGraph];
		currentSubEdges = &subGraph->edges[currentVertex->edgesIndex];

		for ( edgeIndex = 0; edgeIndex < currentVertex->numberOfEdges; edgeIndex++ )
		{
			CL_FormatEdge( GV, &currentGraphVertex->
						edges[currentSubEdges[edgeIndex].indexInGraphVertex],
						graphVertices, labelBuffer);
			strcat(trueLabel, labelBuffer);
			if ( edgeIndex < (currentVertex->numberOfEdges) )	// for all, but the last edge
				strcat(trueLabel, "\\n");							// add new line char in string
		}
	}
}

/*******************************************************************************
FUNCTION NAME: CL_FormatEdge
INPUTS:		PGRAPH_VARIABLES GV
			PGRAPH_EDGE edge, 
			PGRAPH_VERTEX vertices
			char *trueLabel
RETURNS:	void
PURPOSE:	Format the true edge label for the classification lattice.
CALLED BY:	cluster.c: CL_MakeTrueLabel()
*******************************************************************************/

void CL_FormatEdge( PGRAPH_VARIABLES GV, PGRAPH_EDGE edge, 
				   PGRAPH_VERTEX vertices, char *trueLabel )
{
	LABEL label;
	char vLabelBuffer[MAX_TOKEN_LENGTH];
	
#ifdef DEBUG_TRACE
	printf( "%s: CL_FormatEdge()\n", __FILE__ );
#endif
	
	// print edge label
	label = GV->labelList[edge->labelIndex];
	if ( label.labelType == 's' )                               /* string label */
		sprintf( trueLabel, "%s: ", label.content.stringValue );
	else			 											/* numeric label */
		sprintf( trueLabel, "%f: ", label.content.numericValue );

	// print vertex label
	label = GV->labelList[vertices[edge->targetVertexIndex].labelIndex];
	if ( label.labelType == 's' )                               /* string label */
		sprintf( vLabelBuffer, "%s", label.content.stringValue );
	else														/* numeric label */
		sprintf( vLabelBuffer, "%f", label.content.numericValue );

	strcat( trueLabel, vLabelBuffer  );
}


/*******************************************************************************
FUNCTION NAME: IsGraphExhausted
INPUTS:		PGRAPH_VARIABLES GV
RETURNS:	BOOLEAN
PURPOSE:	Check to see if all vertices have label of the form "Sub_n"
CALLED BY:	main.c:	main() (serial)
*******************************************************************************/
BOOLEAN IsGraphExhausted( PGRAPH_VARIABLES GV )
{
	ULONG vertexIndex;
	PGRAPH graph;
	PGRAPH_VERTEX vertices;
	LABEL label;
	
#ifdef DEBUG_TRACE
	printf( "%s: IsGraphExhausted()\n", __FILE__ );
#endif
	
	graph = GV->graph;
	vertices = graph->vertices;
	
	for( vertexIndex = 0; vertexIndex < graph->numberOfVertices; vertexIndex++ )
	{
		label = GV->labelList[vertices[vertexIndex].labelIndex];
		if ( label.labelType == 's' )                               /* string label */
			if (strncmp(label.content.stringValue,			// if label doesn't start with Sub_
						SUB_LABEL_PREFIX, SUB_LABEL_PREFIX_LEN) != 0)
			{
				return FALSE;										// graph is not exhausted
			}
		else														/* numeric label */
			return FALSE;											// -> not Sub_n
	}

	return TRUE;
}

